package com.bergerkiller.bukkit.nolagg;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.bukkit.Location;
import org.bukkit.entity.Entity;
import org.bukkit.entity.ExperienceOrb;
import org.bukkit.entity.Item;
import org.bukkit.inventory.ItemStack;
import org.bukkit.util.Vector;

import com.bergerkiller.bukkit.nolagg.meta.WorldMetaData;

public class WorldStackFormer {
	
	private Object lock = new Object();
	private boolean orbrunning = false;
	private boolean itemrunning = false;
	private final List<Item> items = new ArrayList<Item>();
	private final List<ExperienceOrb> orbs = new ArrayList<ExperienceOrb>();
	private Set<Entity> entitiesToKill = new HashSet<Entity>();
	private Set<Item> itemsToRespawn = new HashSet<Item>();
	public void update(List<Item> itemlist, List<ExperienceOrb> orblist) {
		if (!ItemHandler.formStacks) return;
		if (StackFormer.stackThreshold < 2) return;
		synchronized (lock) {
			if (itemrunning || orbrunning) return;
			for (Entity e : entitiesToKill)  {
			    if (e instanceof Item) WorldMetaData.removeSpawnedItem((Item) e);
				e.remove();
			}
			entitiesToKill.clear();
			for (Item item : itemsToRespawn) {
				ItemHandler.respawnItem(item, new Vector());
			}
			itemsToRespawn.clear();
			if (!orbrunning) {
				orbs.addAll(orblist);
				orbrunning = true;
			}
			if (!itemrunning) {
				items.addAll(itemlist);
				itemrunning = true;
			}
		}		
	}
	
	private void updateOrbs() {
		for (ExperienceOrb orb : orbs) {
			if (entitiesToKill.contains(orb)) continue;
			if (addSameOrbsNear(orbs, orb)) {
				if (near.size() > StackFormer.stackThreshold - 2) {
					for (Entity e : near) {
						if (entitiesToKill.contains(e)) continue;
						//add the experience
						ExperienceOrb to = (ExperienceOrb) e;
						orb.setExperience(orb.getExperience() + to.getExperience());
						entitiesToKill.add(e);
					}						
				}
				near.clear();
			}
		}
		orbs.clear();
	}	
	private void updateItems() {
		ItemStack stack;
		int maxsize;
		for (Item item : items) {
			if (entitiesToKill.contains(item)) continue;
			stack = item.getItemStack();
			maxsize = stack.getMaxStackSize();
			if (stack.getAmount() >= maxsize) continue;
			if (addSameItemsNear(items, item)) {
				if (near.size() > StackFormer.stackThreshold - 2) {
					//addition the items
					Item nearitem;
					ItemStack nearitemstack;
					boolean changed = false;
					for (Entity e : near) {
						if (entitiesToKill.contains(e)) continue;
						nearitem = (Item) e;
						nearitemstack = nearitem.getItemStack();
						if (nearitemstack == null) continue;
						if (nearitemstack.getAmount() >= maxsize) continue;
						int newamount = stack.getAmount() + nearitemstack.getAmount();
						if (newamount <= maxsize) {
							if (newamount == 0) newamount = 1;
							stack.setAmount(newamount);
							entitiesToKill.add(nearitem);
							itemsToRespawn.remove(nearitem);
							changed = true;
						} else {
							//set to max
							stack.setAmount(maxsize);
							//set prev. item
							nearitemstack.setAmount(newamount - maxsize);
							changed = true;
						}
					}
					if (changed && !entitiesToKill.contains(item)) {
						itemsToRespawn.add(item);
					}
				}
				near.clear();
			}
		}
		items.clear();
	}
	
	private static List<Entity> near = new ArrayList<Entity>(StackFormer.stackThreshold - 1);
	private static boolean addSameItemsNear(List<Item> from, Item around) {
		if (around.isDead()) return false;
		Location m = around.getLocation();
		ItemStack stack = around.getItemStack();
		boolean added = false;
		for (Item item : from) {
			if (item.isDead()) continue;
			if (item == around) continue;
			if (canStack(m, item.getLocation())) {
				ItemStack instack = item.getItemStack();
				if (instack.getTypeId() != stack.getTypeId()) {
					continue;
				}
				if (instack.getDurability() != stack.getDurability()) {
					continue;
				}
				added = true;
				near.add(item);
			}
		}
		return added;
	}
	private static boolean addSameOrbsNear(List<ExperienceOrb> from, ExperienceOrb around) {
		if (around.isDead()) return false;
		Location m = around.getLocation();
		boolean added = true;
		for (ExperienceOrb orb : from) {
			if (orb.isDead()) continue;
			if (orb == around) continue;
			if (canStack(m, orb.getLocation())) {
				added = true;
				near.add(orb);
			}
		}
		return added;
	}
	private static boolean canStack(Location l1, Location l2) {
		double d = distance(l1.getX(), l2.getX());
		if (d > StackFormer.stackRadiusSquared) return false;
		d += distance(l1.getZ(), l2.getZ());
		if (d > StackFormer.stackRadiusSquared) return false;
		d += distance(l1.getY(), l2.getY());
		if (d > StackFormer.stackRadiusSquared) return false;
		return true;
	}	
	private static double distance(double d1, double d2) {
		d1 = Math.abs(d1 - d2);
		return d1 * d1;
	}

	public void run() {
		boolean running;
		synchronized (lock) {
			running = itemrunning;
		}
		if (running) {
			updateItems();
			synchronized (lock) {
				itemrunning = false;
			}
		}
		
		synchronized (lock) {
			running = orbrunning;
		}
		if (running) {
			updateOrbs();
			synchronized (lock) {
				orbrunning = false;
			}
		}
	}

}
